# Standard packages
import os
from datetime import datetime

# Additional packages
import cv2
import numpy as np
import torch
import torchvision


# Include file names to save the false positives / false negatives (optional)
class ImageFolderWithFnames(torchvision.datasets.ImageFolder):
    def __getitem__(self, index):
        original_tuple = super(ImageFolderWithFnames, self).__getitem__(index)
        fname = os.path.basename(self.imgs[index][0])
        tuple_with_fname = (original_tuple + (fname,))
        return tuple_with_fname


# Resample image patch
class Resize(object):
    def __init__(self, size):
        assert (isinstance(size, tuple))
        self.size = size

    def __call__(self, sample):
        h, w = self.size
        sample_new = cv2.resize(np.array(sample), (h, w))
        return sample_new[..., np.newaxis]


# Convolutional neural network architecture
class ImgAssistCNN(torch.nn.Module):
    def __init__(self, num_classes=2):
        super(ImgAssistCNN, self).__init__()
        self.features = torch.nn.Sequential(
            torch.nn.Conv2d(1, 8, kernel_size=3, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Conv2d(8, 16, kernel_size=3, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(16, 32, kernel_size=3, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Conv2d(32, 64, kernel_size=3, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2, 2),
            torch.nn.Conv2d(64, 188, kernel_size=3, padding=1),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2, 2)
        )
        self.classifier = torch.nn.Sequential(
            torch.nn.Dropout(0.5),
            torch.nn.Linear(188 * 11 * 11, 64),
            torch.nn.ReLU(inplace=True),
            torch.nn.Dropout(0.5),
            torch.nn.Linear(64, 8),
            torch.nn.ReLU(inplace=True),
            torch.nn.Linear(8, num_classes)
        )

    def forward(self, x):
        x = self.features(x)
        x = x.view(-1, self.num_flat_features(x))
        x = self.classifier(x)
        return x

    @staticmethod
    def num_flat_features(x):
        # all dimensions except the batch dimension
        size = x.size()[1:]
        num_features = 1
        for s in size:
            num_features *= s
        return num_features


# Load TorchScript model and creation timestamp
def load_model(model_dir, model_name):
    print('Loading model ...')
    model_path = model_dir + '/' + model_name
    timestamp = datetime.fromtimestamp(os.path.getmtime(model_path)).strftime('%Y-%m-%d-%H-%M')
    net = torch.jit.load(model_path)
    return net, timestamp
