import csv

# Expected CSV format for Patch Generation Tool:
#  StudyID,UniqueID,ScanID,Ground Truth,DCIS_Positive,DCIS_Negative,IDC,Interface,Cyst,BloodVessel,NormalDuct,Fibroadenoma,Fibrocystic_Change,Bag,Suture,Clip,Adipose,Fibrous,Data Set,IsPositive,HasPositiveLabels,IsValid,# Positive Labels,# Negative Labels
patch_csv_required_num_rows = 18

def patch_generation(csv_filename):

    case_list = {"caseList": []}
    with open(csv_filename) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        ids = []
        for row in csv_reader:
            ids.append(row[1])

        # Skip headers
        ids.pop(0)
        unique_id = list(set(ids))
        unique_id.sort()

        cases = []
        for uid in unique_id:
            cases.append({"uniqueId": uid, "scanId": []})

        csv_file.seek(0)
        for row in csv_reader:
            uid = row[1]
            sid = row[2][1:]
            for case in cases:
                if case["uniqueId"] == uid:
                    case["scanId"].append(int(sid))


        # Only include cases with scans
        for case in cases:
            if case["scanId"]:
                case_list["caseList"].append(case)

    return case_list

# Expected CSV format for Model Generation Tool:
#   Study ID,Unique ID,Scan ID,Ground Truth,numPositiveLabels,numNegativeLabels
model_csv_required_num_rows = 6

def model_generation(csv_filename, split_suspicious):
    nonsuspicious = {}
    suspicious = {}
    with open(csv_filename) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        if split_suspicious:
            nonsuspicious, suspicious = get_cases_split(csv_reader)
        else:
            nonsuspicious = get_all_cases(csv_reader)

    return nonsuspicious, suspicious

def get_all_cases(csv_reader):
    cases = []
    header_skipped = False
    for row in csv_reader:
        if not header_skipped:
            header_skipped = True
            continue
        uid = row[1]
        # Scan ID is S0X, skip the 'S'
        sid = row[2][1:]
        valid, _ = check_margin(row)
        if valid:
            found = False
            for case in cases:
                if case["uniqueId"] == uid:
                    found = True
                    case["scanId"].append(int(sid))
            if not found:
                cases.append({"uniqueId": uid, "scanId": [int(sid)]})
        else:
            print("Invalid margin found: ", row)
    return cases

def get_cases_split(csv_reader):
    non_suspicious_cases = []
    suspicious_cases = []
    header_skipped = False
    for row in csv_reader:
        if not header_skipped:
            header_skipped = True
            continue
        uid = row[1]
        # Scan ID is S0X, skip the 'S'
        sid = row[2][1:]
        valid, suspicious = check_margin(row)

        if valid:
            cases = []
            if suspicious:
                cases = suspicious_cases
            else:
                cases = non_suspicious_cases
            
            found = False
            for case in cases:
                if case["uniqueId"] == uid:
                    found = True
                    case["scanId"].append(int(sid))
            if not found:
                cases.append({"uniqueId": uid, "scanId": [int(sid)]})
        else:
            print("Invalid margin found: ", row)
    
    return non_suspicious_cases, suspicious_cases

# First return value is True if margin is valid, second return value is True if margin is suspicious
def check_margin(csv_row):
    if len(csv_row) < model_csv_required_num_rows:
        return False, False
    ground_truth = csv_row[3]
    margin_is_suspicious = False

    # Shouldn't happen but check anyway
    if "No_Selection" in ground_truth:
        return False, False

    # Skip any margins with Close_DCIS, Close_IDC or Close_ILC
    if "Close" in ground_truth:
        return False, False
    
    if not "Negative" in ground_truth:
        margin_is_suspicious = True

    num_suspicious_labels = int(csv_row[4])
    num_nonsuspicious_labels = int(csv_row[5])

    # Positive ground truth must have suspicious labels
    if margin_is_suspicious and num_suspicious_labels == 0:
        return False, False

    # Negative ground truth doesn't need to have any labels

    # Negative ground truth with suspicious labels, is considered suspicious
    if not margin_is_suspicious and num_suspicious_labels > 0:
        margin_is_suspicious = True

    return True, margin_is_suspicious
    