# Standard packages
import os
import zipfile
import subprocess
import json

# Local modules
from commonTools import wf_oct, common_defs


# Generate file checksum
def generate_hash(filename):
    md5_command = 'md5sum'
    cmd = subprocess.Popen([md5_command, filename], stdout=subprocess.PIPE)
    response = str(cmd.stdout.readline())
    # Result always starts with b', skip these characters
    result = response.split()[0][2:]
    if result[0] == '\\':
        # On windows, Result can have additional \\, skip these characters
        return result.split()[0][2:]
    return result


# Get subfolders for archiving
def get_file_paths(input_dir):
    file_paths = []
    for root, directories, files in os.walk(input_dir):
        for filename in files:
            file_paths.append(os.path.join(root, filename))
    return file_paths


# Unzip list of cases
def extract_data(input_dir, case_list):
    for case in case_list:
        unique_id = case['uniqueId']
        combined_filename = input_dir + '/' + unique_id + '_Combined.zip'
        case_zip_filename = input_dir + '/' + unique_id + '_Native.zip'
        checksum_filename = input_dir + '/' + unique_id + '_Native.txt'
        case_dir = input_dir + '/' + unique_id

        # unzip nested case
        if os.path.exists(combined_filename) and not os.path.exists(case_zip_filename):
            print('Unzipping nested case:', os.path.basename(combined_filename))
            with zipfile.ZipFile(combined_filename, 'r') as zip_case:
                zip_case.extractall(input_dir)

        # unzip case
        if os.path.exists(case_zip_filename) and os.path.exists(checksum_filename):
            if not os.path.isdir(case_dir):
                with open(checksum_filename) as f:
                    case_checksum = generate_hash(case_zip_filename)
                    file_checksum = f.readline().rstrip('\n')
                    if case_checksum == file_checksum:
                        print('Unzipping case:', os.path.basename(case_zip_filename))
                        with zipfile.ZipFile(case_zip_filename, 'r') as zip_case:
                            zip_case.extractall(input_dir)
                    else:
                        print('Checksum failure for case: ', unique_id)


# Read Master Label Set
def read_mls(input_dir, unique_id):
    print('Reading Master Label Set for case %s ...' % unique_id)
    mls_filename = input_dir + '/MLS_' + unique_id + '_labelSet.json'
    mls_checksum_filename = input_dir + '/MLS_' + unique_id + '_labelSet.txt'
    case_checksum_filename = input_dir + '/' + unique_id + '_Native.txt'
    label_data = []
    ground_truth = []

    if os.path.isfile(mls_filename):
        with open(mls_checksum_filename) as f:
            mls_checksum = f.readline().rstrip('\n')

        mls_file_checksum = generate_hash(mls_filename)

        if mls_file_checksum != mls_checksum:
            print('Error: checksum failure for MLS:  ', mls_filename, '. ', mls_file_checksum, ' : ', mls_checksum)
            return label_data, ground_truth

        with open(mls_filename) as f:
            case_label_data = json.load(f)
        file_version = case_label_data['fileVersion']
        if file_version == '1.1.0':
            print('Error: MLS version 1.1.0 not supported')
            return label_data, ground_truth
        elif file_version == '1.2.0':
            upgrade_from_version120(input_dir, unique_id, case_label_data)
        elif file_version == '2.0.0':
            upgrade_from_version200(input_dir, unique_id, case_label_data)

        map_changed_fields(case_label_data)

        label_set_checksum = case_label_data['caseCheckSum'].rstrip('\r')

        with open(case_checksum_filename) as f:
            case_checksum = f.readline().rstrip('\n')

        if label_set_checksum == case_checksum:
            label_data = case_label_data['labels']
            ground_truth = case_label_data['scanGroundTruths']
        else:
            print('Error: checksum failure for case: ', unique_id)
    else:
        print('Warning: Master Label Set not found for case: ', unique_id)

    return label_data, ground_truth


# Upgrade Master Label Set from version 1.2.0
def upgrade_from_version120(input_dir, unique_id, case_label_data):
    print('Upgrading MLS from version 1.2.0 ...')
    for label in case_label_data['labels']:
        scan_id = label['scanId']
        scan_id_str = 'S' + str(scan_id).zfill(2)
        scan_info_filename = input_dir + '/' + unique_id + '/' + scan_id_str + '/scan.json'
        if os.path.exists(scan_info_filename):
            scan_info = wf_oct.ReadScanInfo(scan_info_filename)
            label['y'] -= scan_info.pixDim[1]

# Upgrade Master Label Set from version 2.0.0
def upgrade_from_version200(input_dir, unique_id, case_label_data):
    print('Upgrading MLS from version 2.0.0 ...')
    for label in case_label_data['labels']:
        scan_id = label['scanID']
        bscan = get_wfoct_bscan(input_dir, unique_id, scan_id)
        label['z'] -= bscan.incorrectOffset
        label['z'] += bscan.correctOffset

# In version 2 of the label set, some of the field names were updated
# Re-map them to the old fields here to maintain compatibility
def map_changed_fields(case_label_data):
    print('Mapping new fields to old fields')
    if not 'uniqueId' in case_label_data:
        case_label_data['uniqueId'] = case_label_data['uniqueID']
    if not 'studyId' in case_label_data:
        if 'studyID' in case_label_data:
            case_label_data['studyId'] = case_label_data['studyID']
        else:
            # BAM - 20221205 - Older files didn't include study id
            case_label_data['studyId'] = ""
            case_label_data['studyID'] = ""
    for label in case_label_data['labels']:
        if not 'scanId' in label:
            label['scanId'] = label['scanID']
        if not 'labelId' in label:
            label['labelId'] = label['labelID']
    for groundTruth in case_label_data['scanGroundTruths']:
        if not 'scanId' in groundTruth:
            groundTruth['scanId'] = groundTruth['scanID']

def get_wfoct_bscan(input_dir, unique_id, scan_id):
    scan_id_str = 'S' + str(scan_id).zfill(2)
    bscan_name = input_dir + '/' + unique_id + '/' + scan_id_str + '/OctData_' + str(0).zfill(3) + '.u16'
    scan_data = wf_oct.ReadU16(bscan_name)
    return scan_data

def positive_label(label):
    return label in (common_defs.LabelType.DCIS_Positive, common_defs.LabelType.DCIS_Negative, common_defs.LabelType.IDC, common_defs.LabelType.ILC)


def get_subfolder_from_label(label):
    if label['type'] == common_defs.LabelType.DCIS_Positive:
        return 'DCIS+'
    elif label['type'] == common_defs.LabelType.DCIS_Negative:
        return 'DCIS-'
    elif label['type'] == common_defs.LabelType.IDC:
        return 'IDC'
    elif label['type'] == common_defs.LabelType.Interface:
        return 'Interface'
    elif label['type'] == common_defs.LabelType.Cyst:
        return 'Cyst'
    elif label['type'] == common_defs.LabelType.BloodVessel:
        return 'BloodVessel'
    elif label['type'] == common_defs.LabelType.NormalDuct:
        return 'NormalDuct'
    elif label['type'] == common_defs.LabelType.Fibroadenoma:
        return 'Fibroadenoma'
    elif label['type'] == common_defs.LabelType.Fibrocystic_Change:
        return 'Fibrocystic_Change'
    elif label['type'] == common_defs.LabelType.Bag:
        return 'Bag'
    elif label['type'] == common_defs.LabelType.Suture:
        return 'Suture'
    elif label['type'] == common_defs.LabelType.Clip:
        return 'Clip'
    elif label['type'] == common_defs.LabelType.Adipose:
        return 'Adipose'
    elif label['type'] == common_defs.LabelType.Fibrous:
        return 'Fibrous'
    elif label['type'] == common_defs.LabelType.Skin:
        return 'Skin'
    elif label['type'] == common_defs.LabelType.ILC:
        return 'ILC'
    else:
        return 'Invalid'

def ground_truth_to_string(ground_truth):    
    if ground_truth == common_defs.GroundTruth.Negative:
        return 'Negative'
    elif ground_truth == common_defs.GroundTruth.Positive_IDC:
        return 'Positive_IDC'
    elif ground_truth == common_defs.GroundTruth.Positive_DCIS:
        return 'Positive_DCIS'
    elif ground_truth == common_defs.GroundTruth.Positive_IDC_DCIS:
        return 'Positive_IDC_DCIS'
    elif ground_truth == common_defs.GroundTruth.Close_DCIS:
        return 'Close_DCIS'
    elif ground_truth == common_defs.GroundTruth.Close_IDC:
        return 'Close_IDC'
    elif ground_truth == common_defs.GroundTruth.Positive_ILC:
        return 'Positive_ILC'
    elif ground_truth == common_defs.GroundTruth.Positive_IDC_DCIS_ILC:
        return 'Positive_IDC_DCIS_ILC'
    elif ground_truth == common_defs.GroundTruth.Positive_IDC_ILC:
        return 'Positive_IDC_ILC'
    elif ground_truth == common_defs.GroundTruth.Positive_DCIS_ILC:
        return 'Positive_DCIS_ILC'
    elif ground_truth == common_defs.GroundTruth.Close_ILC:
        return 'Close_ILC'
    else:
        return 'No Selection'