# Standard packages
import json

# Additional packages
import numpy as np


# Read single WF OCT B-scan
class ReadU16:
    def __init__(self, filename):
        with open(filename, 'rb') as file_id:
            self.fileSize = self.fread(file_id, 1, np.uint32)[0]
            self.fileVersion = self.fread(file_id, 1, np.uint32)[0]
            serial_number_length = 10 if self.fileVersion >= 6 else 0
            header_size = 12 * np.dtype(np.uint32).itemsize + np.dtype(np.uint64).itemsize + \
                7 * np.dtype(np.float32).itemsize + serial_number_length
            self.imageFactor = self.fread(file_id, 1, np.uint32)[0]
            self.widthPixels = self.fread(file_id, 1, np.uint32)[0]
            self.heightPixels = self.fread(file_id, 1, np.uint32)[0]
            self.cropStart = self.fread(file_id, 1, np.uint32)[0]
            self.cropEnd = self.fread(file_id, 1, np.uint32)[0]
            self.caseId = self.fread(file_id, 1, np.uint32)[0]
            self.scanId = self.fread(file_id, 1, np.uint32)[0]
            self.sliceNumber = self.fread(file_id, 1, np.uint32)[0]
            self.numCellsAdded = self.fread(file_id, 1, np.uint32)[0]
            self.numCellsExpected = self.fread(file_id, 1, np.uint32)[0]
            self.legacySerialNumber = self.fread(file_id, 1, np.uint64)[0]
            self.widthMm = self.fread(file_id, 1, np.float32)[0]
            self.heightMm = self.fread(file_id, 1, np.float32)[0]
            self.glassHeightMm = self.fread(file_id, 1, np.float32)[0]
            self.xLocationMm = self.fread(file_id, 1, np.float32)[0]
            self.yLocationMm = self.fread(file_id, 1, np.float32)[0]
            self.indexOfRefraction = self.fread(file_id, 1, np.float32)[0]
            self.referenceArmIntensity = self.fread(file_id, 1, np.float32)[0]
            self.serialNumber = self.fread(file_id, serial_number_length, np.uint8)
            self.imageData = self.fread(file_id, int(self.fileSize - header_size), np.uint8)
            self.imageData = np.reshape(self.imageData, (self.heightPixels, self.widthPixels))
            self.incorrectOffset = self.__calculateIncorrectOffset()
            self.correctOffset = self.__calculateCorrectOffset()

    # Read n of elements from file
    @staticmethod
    def fread(file_id, num_elements, dtype):
        # Equivalent to Matlab fread function
        if dtype is np.str_:
            # WARNING: assuming 8-bit ASCII for np.str_!
            dtype = np.uint8
        data_array = np.fromfile(file_id, dtype, num_elements)
        return data_array

    # Version 2.0.0 of label files have incorrect offset, calculate correction factors here
    def __calculateIncorrectOffset(self):
        return self.cropStart * (self.heightPixels/self.heightMm) / self.indexOfRefraction

    def __calculateCorrectOffset(self):
        return self.cropStart * (self.heightMm/self.heightPixels) / self.indexOfRefraction


# Get scan config info from scan.json
class ReadScanInfo:
    def __init__(self, filename):
        with open(filename) as f:
            data = json.load(f)
        self.numOctSlices = data['numOctSlices']
        # check for file version: OTIS 2.1 vs Selene
        version = data['swVersion']
        # verison 0.1.1.0 corresponds to converted OTIS20 data
        if version == '0.1.1.0':
            self.pixDim = self.read_otis21_info(data)
        elif version == '2.2.0.0':
            self.pixDim = self.read_otis21_info(data)
        elif version == '2.3.0.0':
            self.pixDim = self.read_otis21_info(data)
        else:
            self.pixDim = self.read_selene_info(data)

    @staticmethod
    def read_otis21_info(data):
        config = data['scanConfig']
        oct_cell_num_pixels = np.array(config['octCellNumPixels'])
        oct_cell_size_mm = np.array(config['octCellSizeMm'])
        pix_dim = oct_cell_size_mm / oct_cell_num_pixels
        return pix_dim

    @staticmethod
    def read_selene_info(data):
        config = data['scanAppConfig']
        oct_cell_num_pixels = config['octCellNumPixels']
        oct_cell_size_mm = config['octCellSizeMm']
        pix_dim = np.array([oct_cell_size_mm['x'] / oct_cell_num_pixels['x'],
                            oct_cell_size_mm['y'] / oct_cell_num_pixels['y'],
                            oct_cell_size_mm['z'] / oct_cell_num_pixels['z']])
        return pix_dim
