#!/usr/bin/python3


# Standard packages
import sys
import os
import glob
import json
import csv

# Local Modules
from commonTools import common_defs, utils
   
# Print program usage
def usage(progname):
    print('ImgAssist Labelset to CSV Converter\n\nUsage: %s %s' % (progname, '<input_folder> <out_file>'))

    
# Read Master Label Set
def read_mls(mls):
    mls_filename = mls + '.json'
    mls_checksum_filename = mls + '.txt'
    label_data = []
    ground_truth = []

    if os.path.isfile(mls_filename):
        with open(mls_checksum_filename) as f:
            mls_checksum = f.readline().rstrip('\n')

        mls_file_checksum = utils.generate_hash(mls_filename)

        if mls_file_checksum != mls_checksum:
            print('Checksum failure for MLS:  ', mls_filename, '. ', mls_file_checksum, ' : ', mls_checksum)
            return label_data, ground_truth

        with open(mls_filename) as f:
            case_label_data = json.load(f)

        file_version = case_label_data['fileVersion']
        if file_version == '1.1.0':
            print('Error: MLS version 1.1.0 not supported')
            return label_data, ground_truth
        elif file_version == '2.0.0' or file_version == '2.0.1':
            for label in case_label_data['labels']:
                # Version 2.0.0 renamed the following fields, so create mapping here to the old terms
                label['scanId'] = label['scanID']
                label['labelId'] = label['labelID']
            for groundTruth in case_label_data['scanGroundTruths']:
                groundTruth['scanId'] = groundTruth['scanID']

        if case_label_data['fileVersion'] != "1.2.0":
            print('New fileversion ', case_label_data['fileVersion'], ', for mls ', mls)

        label_data = case_label_data['labels']
        ground_truth = case_label_data['scanGroundTruths']
    else:
        print('Master Label Set not found for case: ', mls)

    return label_data, ground_truth


# Extract data
def extract_data(mls_list, out_filename):
    with open(out_filename, mode='w', newline='') as out_file:
        writer = csv.writer(out_file, delimiter=',', quoting=csv.QUOTE_MINIMAL)
        labels = ['UniqueID','ScanID','Ground Truth']
        for label_type in common_defs.LabelType:
            labels.append(label_type.name)

        writer.writerow(labels)
        
        for mls in mls_list:
            label_data, ground_truth = read_mls(mls)
            if not (label_data or ground_truth):
                continue

            scan_ids = []
            for scan in range(0,len(ground_truth)):
                scan_id = ground_truth[scan]['scanId']
                scan_ids.append(scan_id)
                num_case_pos_labels = 0
                num_case_neg_labels = 0
                num_case_exclude_labels = 0

            unique_id = os.path.basename(mls)[4:-9]

            for scan in range(len(ground_truth)):
                ground_truth_str = common_defs.groundTruthStrings[ground_truth[scan]['groundTruth']]
                if ground_truth[scan]['groundTruth'] > common_defs.GroundTruth.Close_IDC:
                    print("MLS %s has new ground truth %s" %(mls, ground_truth_str))
                scan_labels=[unique_id, 'S'+str(scan_ids[scan]).zfill(2), ground_truth_str]
                for label_type in common_defs.LabelType:
                    num_labels = 0
                    for label in label_data:
                        idx = label['type']
                        scan_id = label['scanId']
                        if scan_id == scan_ids[scan] and label_type.value == idx:
                            num_labels += 1
                            if idx == common_defs.LabelType.DCIS_Positive or idx == common_defs.LabelType.DCIS_Negative or idx == common_defs.LabelType.IDC or idx == common_defs.LabelType.ILC:
                                num_case_pos_labels += 1
                                if idx == common_defs.LabelType.ILC:
                                    print("MLS %s has new ILC label" %(mls))
                            elif idx != common_defs.LabelType.Exclude_Margin:
                                num_case_neg_labels += 1
                                if idx == common_defs.LabelType.Skin:
                                    print("MLS %s has new skin label" %(mls))
                            else:
                                num_case_exclude_labels += 1
                                
                    scan_labels.append(num_labels)
                writer.writerow(scan_labels)
            
            if num_case_exclude_labels > 0:
                print("MLS %s has %d exclude labels" % (mls, num_case_exclude_labels))
                                

###
def main(argc, argv):
    # check input arguments
    if argc != 3:
        usage(os.path.basename(argv[0]))
        sys.exit(2)

    input_dir = argv[1]
    if not os.path.exists(input_dir):
        print('Error: input folder not found, exiting ...')
        sys.exit(0)
    out_filename = argv[2]

    mls_list = []
    file_list = glob.glob(input_dir + '/**/MLS*.json', recursive=True)
    for filename in file_list:
        # Some cases have old MLS files with text appended to the end, exclude these
        if "labelSet.json" in filename:
            mls_list.append(os.path.splitext(filename)[0])
            #print(filename)

    # read and unzip input cases
    extract_data(mls_list, out_filename)