# Standard packages
import glob
import os
import sys
import random
from shutil import copy
from zipfile import ZipFile

# Local modules
from commonTools import common_defs
from commonTools.utils import get_file_paths
from modelGenerationTool import constants


# Copy patches from input subfolders to model training folder
def copy_patches(cases, input_dir, subfolders, output_dir):
    for case in cases:
        unique_id = case['uniqueId']
        for scan_id in case['scanId']:
            scan_id_str = 'S' + str(scan_id).zfill(2)
            for folder in subfolders:
                for filename in glob.glob(
                        input_dir + '/' + unique_id + '/' + folder + '/' + unique_id + '_' + scan_id_str + '*'):
                    copy(filename, output_dir)


# Delete random patches from negative class
def delete_random_patches(current_dir, num_pos_files, num_neg_files, pos_scale_fac):
    files = glob.glob(current_dir + '/' + common_defs.nonSuspiciousSubfolder + '/*')
    num_neg_files_to_keep = num_pos_files * pos_scale_fac
    while num_neg_files_to_keep < num_neg_files:
        filepath = random.choice(files)
        if os.path.isfile(filepath):
            os.remove(filepath)
            num_neg_files -= 1


# Prepare training/validation/testing data folders for model generation
def copy_data(input_dir, train_cases_pos, train_cases_neg, val_cases_pos, val_cases_neg, test_cases,
              train_dir, val_dir, test_dir, pos_scale_fac):
    print('Preparing folders ...')
    for filename in os.listdir(train_dir + '/' + common_defs.suspiciousSubfolder):
        os.remove(train_dir + '/' + common_defs.suspiciousSubfolder + '/' + filename)
    for filename in os.listdir(train_dir + '/' + common_defs.nonSuspiciousSubfolder):
        os.remove(train_dir + '/' + common_defs.nonSuspiciousSubfolder + '/' + filename)
    for filename in os.listdir(val_dir + '/' + common_defs.suspiciousSubfolder):
        os.remove(val_dir + '/' + common_defs.suspiciousSubfolder + '/' + filename)
    for filename in os.listdir(val_dir + '/' + common_defs.nonSuspiciousSubfolder):
        os.remove(val_dir + '/' + common_defs.nonSuspiciousSubfolder + '/' + filename)

    print('Copying training data ...')
    copy_patches(train_cases_pos, input_dir, common_defs.posSubfolders,
                 train_dir + '/' + common_defs.suspiciousSubfolder)
    copy_patches(train_cases_neg, input_dir, common_defs.negSubfolders + [common_defs.randomSubfolder],
                   train_dir + '/' + common_defs.nonSuspiciousSubfolder)
    num_train_pos_files = len(os.listdir(train_dir + '/' + common_defs.suspiciousSubfolder))
    num_train_neg_files = len(os.listdir(train_dir + '/' + common_defs.nonSuspiciousSubfolder))
    if num_train_pos_files == 0:
        print('Error: empty suspicious training folder, exiting ...')
        sys.exit(0)
    if num_train_neg_files == 0:
        print('Error: empty non-suspicious training folder, exiting ...')
        sys.exit(0)
    train_ratio = num_train_neg_files // num_train_pos_files

    print('Copying validation data ...')
    copy_patches(val_cases_pos, input_dir, common_defs.posSubfolders,
                 val_dir + '/' + common_defs.suspiciousSubfolder)
    copy_patches(val_cases_neg, input_dir, common_defs.negSubfolders + [common_defs.randomSubfolder],
                 val_dir + '/' + common_defs.nonSuspiciousSubfolder)
    num_val_pos_files = len(os.listdir(val_dir + '/' + common_defs.suspiciousSubfolder))
    num_val_neg_files = len(os.listdir(val_dir + '/' + common_defs.nonSuspiciousSubfolder))
    if num_val_pos_files == 0:
        print('Error: empty suspicious validation folder, exiting ...')
        sys.exit(0)
    if num_val_neg_files == 0:
        print('Error: empty non-suspicious validation folder, exiting ...')
        sys.exit(0)
    val_ratio = num_val_neg_files // num_val_pos_files

    # using all available samples
    if pos_scale_fac == 0:
        pos_scale_fac = train_ratio if train_ratio < val_ratio else val_ratio

    print('Processing training and validation datasets ...')
    delete_random_patches(train_dir, num_train_pos_files, num_train_neg_files, pos_scale_fac)
    delete_random_patches(val_dir, num_val_pos_files, num_val_neg_files, pos_scale_fac)

    for filename in os.listdir(test_dir + '/' + common_defs.suspiciousSubfolder):
        os.remove(test_dir + '/' + common_defs.suspiciousSubfolder + '/' + filename)
    for filename in os.listdir(test_dir + '/' + common_defs.nonSuspiciousSubfolder):
        os.remove(test_dir + '/' + common_defs.nonSuspiciousSubfolder + '/' + filename)

    print('Copying test data ...')
    copy_patches(test_cases, input_dir, common_defs.posSubfolders, test_dir + '/' + common_defs.suspiciousSubfolder)
    all_neg_folders = common_defs.negSubfolders
    all_neg_folders.append(common_defs.randomSubfolder)
    copy_patches(test_cases, input_dir, all_neg_folders, test_dir + '/' + common_defs.nonSuspiciousSubfolder)


# Archive data sets
def archive_data(input_dir, save_dir):
    curr_dir = os.getcwd()
    os.chdir(input_dir)
    for subfolder in ['Training', 'Validation']:
        file_paths = get_file_paths('./' + subfolder)
        zip_file_path = input_dir + '/' + subfolder + '.zip'
        zip_file = ZipFile(zip_file_path, 'w')
        with zip_file:
            for filename in file_paths:
                zip_file.write(filename)
        copy(zip_file_path, save_dir)
        os.remove(zip_file_path)
    os.chdir(curr_dir)
