# Standard packages
import os
from shutil import rmtree

# Additional packages
import matplotlib.pyplot as plt
import numpy as np
from sklearn import metrics
from sklearn.metrics import auc, matthews_corrcoef, f1_score
import torch
import torchvision
from torchvision import transforms
import time

# Local modules
from commonTools import common_defs, imgassist
from modelGenerationTool import constants

# Default initialization of model weights to ensure reproducibility
def weights_init(m):
    if isinstance(m, torch.nn.Conv2d):
        torch.nn.init.xavier_uniform_(m.weight)
        torch.nn.init.zeros_(m.bias)


# Train and save ImgAssist model
def train_model(train_dir, val_dir, num_epochs, num_epoch_stop, weight_reduction_coeff, save_dir, model_name, learning_rate, weight_decay, training_batch_size):
    # remove randomness from initialization
    st = time.time()
    np.random.seed(0)
    torch.manual_seed(0)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    net = imgassist.ImgAssistCNN().cuda()
    net.apply(weights_init)

    num_pos_samples = len([name for name in os.listdir(train_dir + '/' + common_defs.suspiciousSubfolder)])
    num_neg_samples = len([name for name in os.listdir(train_dir + '/' + common_defs.nonSuspiciousSubfolder)])
    pos_weight = torch.tensor(weight_reduction_coeff * num_neg_samples / num_pos_samples).cuda()

    # binary cross-entropy is used for 2-class architecture only
    criterion = torch.nn.BCEWithLogitsLoss(pos_weight=pos_weight)
    optimizer = torch.optim.Adam(net.parameters(), lr=learning_rate, weight_decay=weight_decay)

    # directory to save the training iterations
    checkpoint_dir = save_dir + '/Checkpoints/'
    if os.path.exists(checkpoint_dir):
        rmtree(checkpoint_dir)
    os.mkdir(checkpoint_dir)

    # patch transformation
    data_transform = transforms.Compose([
        # convert PNG to 1-channel (grayscale) image
        transforms.Grayscale(),
        # downsample patch
        imgassist.Resize((constants.patchSize, constants.patchSize)),
        # transfom patch data to a PyTorch tensor with values [0:1]
        transforms.ToTensor(),
        # normalize values to 0.5 mean and 0.5 stddev (corresponds to range [-1:1])
        transforms.Normalize([constants.dataMean], [constants.dataStdDev])])

    datasets = {'training': torchvision.datasets.ImageFolder(train_dir, data_transform),
                'validation': torchvision.datasets.ImageFolder(val_dir, data_transform)}
    dataloaders = {
        'training': torch.utils.data.DataLoader(datasets['training'], batch_size=training_batch_size, shuffle=True, num_workers=16),
        'validation': torch.utils.data.DataLoader(datasets['validation'], batch_size=training_batch_size, shuffle=False, num_workers=16)}
    dataset_size = {'training': len(datasets['training']), 'validation': len(datasets['validation'])}
    print('Datasets:', dataset_size)
    class_names = datasets['training'].classes
    print('Classes:', class_names)

    epochs_no_improve = 0
    best_epoch = 0
    min_val_loss = np.Inf
    t_acc = []
    v_acc = []
    t_loss = []
    v_loss = []
    training_acc = []
    validation_acc = []


    # training loop
    for epoch in range(num_epochs):
        running_loss = 0.0
        val_loss = 0.0
        for phase in ['training', 'validation']:
            if phase == 'training':
                net.train()
            else:
                net.eval()

            for i, data in enumerate(dataloaders[phase]):
                # get the inputs        
                inputs, labels = data
                # conversion to float is required when using BCE
                inputs = inputs.cuda()
                labels = labels.float().cuda()

                # zero the parameter gradients       
                optimizer.zero_grad()

                # forward + backward + optimize
                outputs = net(inputs)[:, 1]
                loss = criterion(outputs, labels)
                if phase == 'training':
                    outputs_t = net(inputs)
                    loss.backward()
                    optimizer.step()
                    running_loss += loss.item()
                    _, pred_t = torch.max(outputs_t, dim = 1)
                    correct_t = torch.sum(pred_t == labels).item()
                    total_t = labels.size(0)
                    t_acc.append(100 * correct_t / total_t)


                    # print running loss after every nth batch
                    if i % constants.printLoss == constants.printLoss - 1:
                        print('\rEpoch: [%d/%d] Progress: [%5d] ' %(epoch + 1, num_epochs, i + 1), end='')

                else:
                    outputs_v = net(inputs)
                    val_loss += loss.item()
                    _ , pred_v = torch.max(outputs_v, dim = 1)
                    correct_v = torch.sum(pred_v ==labels).item()
                    total_v = labels.size(0)
                    v_acc.append(100 * correct_v / total_v)

            train_loss = running_loss / len(dataloaders['training'])
            val_loss = val_loss / len(dataloaders['validation'])

        train_acc_mean = np.mean(t_acc)
        training_acc.append(train_acc_mean)
        t_loss.append(train_loss)

        val_acc_mean = np.mean(v_acc)
        validation_acc.append(val_acc_mean)
        v_loss.append(val_loss)

        print('\nTraining loss: %.3f' % (train_loss))
        print('Validation loss: %.3f' % (val_loss))
        print(f'Training Accuracy: {train_acc_mean:.3f}%' )
        print(f'Validation Accuracy: {val_acc_mean: .3f}% \n')


        # save the result of training epoch
        torch.save(net.state_dict(),
                   checkpoint_dir + 'CP{}.pth'.format(epoch + 1))
        if val_loss < min_val_loss:
            epochs_no_improve = 0
            best_epoch = epoch + 1
            min_val_loss = val_loss
            print('Improvement detected on epoch %d ...\n' % (epoch +1))
        else:
            epochs_no_improve += 1
            # check early stopping condition
            if epochs_no_improve == num_epoch_stop:
                print('Stop at epoch no.', best_epoch)
                # loading the best performing model
                model = checkpoint_dir + 'CP' + str(best_epoch) + '.pth'
                net.load_state_dict(torch.load(model))
                break
    #save_training_data(training_acc,validation_acc,t_loss,v_loss,save_dir) #comment out if not debugging
    et = time.time()
    elapsed_time = et - st

    print(f'Training Time: {elapsed_time:.1f}s')



    net.eval()
    scripted_module = torch.jit.script(net)
    out_filepath = save_dir + '/' + model_name
    scripted_module.save(out_filepath)
    print('Model saved')

    # cleanup
    rmtree(checkpoint_dir)


# Evaluate generated model
def eval_model(test_dir, save_dir, model_name, confidence_step, report_name='ModelValidationReport.txt',
               roc_name='ROC.pdf', pr_name = 'PR.pdf'):
    st = time.time()
    data_transform = transforms.Compose([
        transforms.Grayscale(),
        imgassist.Resize((constants.patchSize, constants.patchSize)),
        transforms.ToTensor(),
        transforms.Normalize([constants.dataMean], [constants.dataStdDev])
    ])

    net, model_timestamp = imgassist.load_model(save_dir, model_name)

    datasets = {'test': imgassist.ImageFolderWithFnames(test_dir, data_transform)}
    dataloaders = {'test': torch.utils.data.DataLoader(datasets['test'], batch_size=256, shuffle=False,
                                                       num_workers=16)}
    dataset_size = {'test': len(datasets['test'])}
    print('Datasets:', dataset_size)
    class_names = datasets['test'].classes
    print('Classes:', class_names)

    sensitivity = [1.0]
    specificity = [0.0]
    precision = [0.0]
    f1 = [0.0]
    mcc = [0.0]

    for confidence in np.arange(0.5, 1.0, confidence_step):
        t = torch.tensor([confidence]).cuda()
        iter_true_positives = 0
        iter_false_positives = 0
        iter_true_negatives = 0
        iter_false_negatives = 0
        num_patches = 0

        # testing loop
        with torch.no_grad():
            net.eval()
            num_true_positives = 0
            num_false_positives = 0
            num_true_negatives = 0
            num_false_negatives = 0

            test_pred, test_target = [],[]

            for data in dataloaders['test']:
                images, labels, fnames = data
                images = images.cuda()
                outputs = net(images)
                # convert outputs to probabilities
                outputs_softmax = torch.softmax(outputs,dim=1)
                prob, predicted = torch.max(outputs_softmax.data, 1)
                # long() converts bool to int
                mask = (prob >= t).long() * 1
                predicted *= mask
                predicted = predicted.cpu()

                for i in range(len(predicted)):

                    test_pred.append(predicted[i])
                    test_target.append(labels[i])

                for i in range(labels.size(0)):
                    if predicted[i] == common_defs.suspicious:
                        if predicted[i] == labels[i]:
                            num_true_positives += 1
                        else:
                            num_false_positives += 1
                    elif predicted[i] == common_defs.nonSuspicious:
                        if predicted[i] == labels[i]:
                            num_true_negatives += 1
                        else:
                            num_false_negatives += 1

                num_patches += labels.size(0)
                print('\rClassification threshold: %4.3f => processed %d / %d patches' %
                      (confidence, num_patches, len(datasets['test'])), end='')

            iter_true_positives += num_true_positives
            iter_false_positives += num_false_positives
            iter_true_negatives += num_true_negatives
            iter_false_negatives += num_false_negatives

        iter_sensitivity = iter_true_positives / (iter_true_positives + iter_false_negatives)
        iter_specificity = iter_true_negatives / (iter_true_negatives + iter_false_positives)
        prec_den = iter_true_positives + iter_false_positives
        iter_precision = iter_true_positives / prec_den if prec_den != 0 else 0 #non-zero denominator check
        
        sensitivity.append(iter_sensitivity)
        specificity.append(iter_specificity)
        precision.append(iter_precision)
        f1.append(f1_score(test_target,test_pred, average = 'macro'))
        mcc.append(matthews_corrcoef(test_target,test_pred))

    # Add values for confidence threshold of 1
    sensitivity.append(0.0)
    specificity.append(1.0)
    precision.append(1.0)
    f1.append(0.0)
    mcc.append(0.0)
    et = time.time()

    elapsed_test_time = et - st
    print(f'\nInferencing time of test set: {elapsed_test_time:.1f}s')

    print('\nGenerating report ...')
    generate_report(np.asarray(sensitivity), np.asarray(specificity),np.asarray(precision), np.asarray(f1), confidence_step, save_dir,
                    model_name, model_timestamp, report_name, roc_name, pr_name,mcc)



# Generate performance evaluation report
def generate_report(sensitivity, specificity,precision, f1, confidence_step, save_dir, model_name, model_timestamp,
                    report_name, roc_name, pr_name,mcc):
    model_filepath = save_dir + '/' + model_name
    report_filepath = save_dir + '/' + report_name
    with open(report_filepath, 'w') as f:
        f.write('ImgAssist Model Generation Tool v. %s\n\n' % constants.version)
        f.write('### MODEL ###\n')
        f.write('Model: %s\n' % model_filepath)
        f.write('Version: %s\n\n' % model_timestamp)

        f.write('### CLASSIFICATION THRESHOLD: SENSITIVITY / SPECIFICITY / PRECISION / F1-SCORE / MCC ###\n')
        num_entries = len(sensitivity)
        # start with threshold 0.5 and ignore the first entry

        for i in range(num_entries - 1):
            f.write(
                '%4.3f: %.3f / %.3f / %.3f / %.3f / %.3f\n' % (0.5 + i * confidence_step, sensitivity[i + 1], specificity[i + 1], precision[i+1], f1[i + 1], mcc[i+1]))




        Auc = auc(1.0 - specificity, sensitivity)
        f.write('\n### AUC ###\n')
        f.write('AUC value: %.3f\n' % Auc)
        plt.show()

        AUPRC = auc(sensitivity,precision)
        f.write('\n### AUPRC ###\n')
        f.write('AUPRC value: %.3f\n' % AUPRC)



    # plot ROC (optional)
    if roc_name:
        print('Plotting ROC curve ...')
        auc_str = 'AUC: ' + '{:.3f}'.format(Auc)

        # diagonal line
        x_vals = []
        y_vals = []
        xt = 0.0
        while xt <= 1.0:
            x_vals.append(xt)
            y_vals.append(xt)
            xt += 0.1

        # plot AUC value at defined position
        plt.text(0.1, 0.85, auc_str)
        # plot curve
        plt.xlabel('False positive rate')
        plt.ylabel('Sensitivity')
        plt.xlim(0.0, 1.0)
        plt.ylim(0.0, 1.0)
        plt.plot(1.0 - specificity, sensitivity)
        plt.plot(x_vals, y_vals, dashes=[3, 1])

        roc_filepath = save_dir + '/' + roc_name
        plt.savefig(roc_filepath)

    if pr_name:
        print('Plotting Precision-Recall curve ...')
        auprc_str = 'AUPRC: ' + '{:.3f}'.format(AUPRC)

        # diagonal line
        x_vals = []
        y_vals = []
        xt = 0.0
        while xt <= 1.0:
            x_vals.append(xt)
            y_vals.append(xt)
            xt += 0.1

        # plot AUPRC value at defined position
        plt.figure(2)
        plt.text(0.1, 0.85, auprc_str)
        # plot curve
        plt.xlabel('Recall (Sensitivity)')
        plt.ylabel('Precision')
        plt.xlim(0.0, 1.0)
        plt.ylim(0.0, 1.0)
        plt.plot(sensitivity, precision)

        pr_filepath = save_dir + '/' + pr_name
        plt.savefig(pr_filepath)

#Generate training-validation loss and accuracy curves for debugging and internal model performance reporting
def save_training_data(training_acc,validation_acc,t_loss,v_loss,save_dir):
    print('Debug Mode: Saving training-validation curves ... \n')
    plt.figure(figsize=(20, 10))
    plt.title('Train-Validation Accuracy')
    plt.plot(training_acc, label='Train')
    plt.plot(validation_acc, label='Validation')
    plt.xlabel('num_epochs', fontsize=12)
    plt.ylabel('Loss', fontsize=12)
    plt.legend(loc='best')
    plt.savefig(save_dir + '/' + 'Train-Validation Accuracy.pdf')

    plt.figure(2)
    plt.title('Train-Validation Loss')
    plt.plot(t_loss, label='train')
    plt.plot(v_loss, label='validation')
    plt.xlabel('num_epochs', fontsize=12)
    plt.ylabel('Loss', fontsize=12)
    plt.legend(loc='best')
    plt.savefig(save_dir + '/' + 'Train-Validation Loss.pdf')