# Standard packages
import json
import os
import shutil
import sys

# Additional packages
import torch

# Local modules
from modelGenerationTool import constants, model_application, data_manager
from commonTools import parse_csv

#Change default share stategy 
torch.multiprocessing.set_sharing_strategy('file_system')

# Print program usage
def usage(progname, version):
    options = 'Options:\n\ninit\n\tGenerate datasets\ntrain\n\tTrain model\ntest\n\tValidate model'
    print('ImgAssist Model Generation Tool v. %s\n\nUsage: %s [--option] %s\n\n%s' %
          (version, progname, '<input_params.json>', options))


# Read input parameters
def read_params(filename, mode):
    global input_data_dir
    global train_data_dir
    global train_dir
    global val_dir
    global test_dir
    global save_dir
    global training_set_filename
    global validation_set_filename
    global test_set_filename
    global num_epochs
    global num_epoch_stop
    global model_name
    global pos_scale_fac
    global weight_reduction_coeff
    global confidence_step
    global learning_rate
    global weight_decay
    global training_batch_size

    with open(filename) as f:
        data = json.load(f)

    tool_name = data['tool']
    if tool_name != constants.tool:
        print('Error: input parameters incompatible with tool ...')
        sys.exit(0)

    input_version = data['version']
    if constants.version != input_version:
        print('Error: tool version is not compatible with input ...')
        sys.exit(0)

    input_data_dir = data['inputDataFolder']
    if not os.path.exists(input_data_dir):
        print('Error: input data folder not found, exiting ...')
        sys.exit(0)

    train_data_dir = data['trainingDataFolder']
    if not os.path.exists(train_data_dir):
        print('Error: training data folder not found, exiting ...')
        sys.exit(0)
        
    train_dir = ''
    val_dir = ''
    test_dir = ''

    train_dir = train_data_dir + '/Training'
    val_dir = train_data_dir + '/Validation'
    test_dir = train_data_dir + '/Testing'

    if mode == constants.trainMode or mode == constants.testMode:
        if not os.path.exists(train_dir):
            print('Error: training folder not found, exiting ...')
            sys.exit(0)

        if not os.path.exists(val_dir):
            print('Error: validation folder not found, exiting ...')
            sys.exit(0)

        if not os.path.exists(test_dir):
            print('Error: testing folder not found, exiting ...')
            sys.exit(0)

    if mode == constants.initMode or mode == constants.execMode:
        print('Allocating data folders ...')
        if os.path.isdir(train_dir):
            shutil.rmtree(train_dir)
        os.mkdir(train_dir)
        os.mkdir(train_dir + '/Suspicious')
        os.mkdir(train_dir + '/Non-suspicious')

        if os.path.isdir(val_dir):
            shutil.rmtree(val_dir)
        os.mkdir(val_dir)
        os.mkdir(val_dir + '/Suspicious')
        os.mkdir(val_dir + '/Non-suspicious')

        if os.path.isdir(test_dir):
            shutil.rmtree(test_dir)
        os.mkdir(test_dir)
        os.mkdir(test_dir + '/Suspicious')
        os.mkdir(test_dir + '/Non-suspicious')

    num_epochs = data['numEpochs']
    if num_epochs <= 0 and mode == constants.trainMode:
        print('Error: invalid number of training epochs, exiting ...')
        sys.exit(0)

    num_epoch_stop = data['epochsNoImprove']
    if num_epoch_stop <= 0 and mode == constants.trainMode:
        print('Error: invalid number of epochs without improvement, exiting ...')
        sys.exit(0)

    pos_scale_fac = data['suspiciousClassScaleFactor']
    if pos_scale_fac < 0:
        print('Error: invalid suspicious class scale factor, exiting ...')
        sys.exit(0)

    weight_reduction_coeff = data['weightReductionCoefficient']
    if weight_reduction_coeff <= 0.0:
        print('Error: invalid weight reduction coefficient, exiting ...')
        sys.exit(0)

    confidence_step = data['classificationConfidenceStep']
    if confidence_step <= 0.0 or confidence_step > 0.5:
        print('Error: invalid classification confidence step, exiting ...')
        sys.exit(0)

    learning_rate = data['learningRate']
    if learning_rate <= 0.0 or learning_rate >= 1.0:
        print('Error: invalid learning rate, exiting ...')
        sys.exit(0)

    weight_decay = data['weightDecay']
    if weight_decay < 0:
        print('Error: invalid weight decay, exiting ...')
        sys.exit(0)

    training_batch_size = data['trainingBatchSize']
    if training_batch_size <= 0:
        print('Error: invalid batch size, exiting ...')
        sys.exit(0)

    training_set_filename = data['trainingSetFilename']
    if mode == constants.initMode or mode == constants.trainMode or mode == constants.execMode:
        if not training_set_filename:
            print('Error: training set filename not specified, exiting ...')
            sys.exit(0)
        elif not os.path.exists(training_set_filename):
            print('Error: training set file not found, exiting ...')
            sys.exit(0)

    validation_set_filename = data['validationSetFilename']
    if mode == constants.initMode or mode == constants.trainMode or mode == constants.execMode:
        if not validation_set_filename:
            print('Error: validation set filename not specified, exiting ...')
            sys.exit(0)
        elif not os.path.exists(validation_set_filename):
            print('Error: validation set file not found, exiting ...')
            sys.exit(0)

    test_set_filename = data['testSetFilename']
    if mode == constants.initMode or constants.testMode or mode == constants.execMode:
        if not test_set_filename:
            print('Error: test set filename not specified, exiting ...')
            sys.exit(0)
        elif not os.path.exists(test_set_filename):
            print('Error: test set file not found, exiting ...')
            sys.exit(0)

    save_dir = data['destinationFolder']
    if not os.path.exists(save_dir):
        print('Error: destination folder not found, exiting ...')
        sys.exit(0)

    model_name = data['modelFileName']
    if len(model_name) == 0:
        print('Error: invalid model file name, exiting ...')
        sys.exit(0)
        

def load_csv(mode):
    global train_scans_pos
    global train_scans_neg
    global val_scans_pos
    global val_scans_neg
    global test_scans
    if mode == constants.initMode or mode == constants.trainMode or mode == constants.execMode:
        train_scans_neg, train_scans_pos = parse_csv.model_generation(training_set_filename, True)
        val_scans_neg, val_scans_pos = parse_csv.model_generation(validation_set_filename, True)
    
    if mode == constants.initMode or mode == constants.testMode or mode == constants.execMode:
        test_scans, _ = parse_csv.model_generation(test_set_filename, False)

###
def main(argc, argv):
    # check input arguments
    if argc != 2 and argc != 3:
        usage(os.path.basename(argv[0]), constants.version)
        sys.exit(2)

    mode = 0
    if argc == 2:
        param_file = argv[1]
        mode = constants.execMode
    else:
        param_file = argv[2]
        if argv[1] == '--init':
            mode = constants.initMode
        if argv[1] == '--train':
            mode = constants.trainMode
        elif argv[1] == '--test':
            mode = constants.testMode
    if not mode:
        usage(os.path.basename(argv[0]), constants.version)
        sys.exit(2)

    # only CUDA-enabled version is supported
    if not torch.cuda.is_available():
        print('Error: CUDA support not available, exiting ...')
        sys.exit(0)

    # read input parameters
    if not os.path.exists(param_file):
        print('Error: input parameter file not found, exiting ...')
        sys.exit(0)

    read_params(param_file, mode)
    load_csv(mode)

    if mode == constants.initMode or mode == constants.trainMode or mode == constants.execMode:
        print('Found %d suspicious and %d nonsuspicious cases in training set' % (len(train_scans_pos), len(train_scans_neg)))
        print('Found %d suspicious and %d nonsuspicious cases in validation set' % (len(val_scans_pos), len(val_scans_neg)))
    
    if mode == constants.initMode or mode == constants.testMode or mode == constants.execMode:
        print('Found %d cases in test set' % (len(test_scans)))

    if mode == constants.initMode or mode == constants.execMode:
        # copy relevant data to the processing folders and ensure a 50/50 split
        data_manager.copy_data(input_data_dir, train_scans_pos, train_scans_neg, val_scans_pos, val_scans_neg,
                               test_scans, train_dir, val_dir, test_dir, pos_scale_fac)

    if mode == constants.trainMode or mode == constants.execMode:
        # train and save the model
        print('Starting model training ...')
        model_application.train_model(train_dir, val_dir, num_epochs, num_epoch_stop, weight_reduction_coeff,
                                      save_dir, model_name, learning_rate, weight_decay, training_batch_size)
        # zip and save training/validation data for further reference
        print('Archiving data ...')
        data_manager.archive_data(train_data_dir, save_dir)

    if mode == constants.testMode or mode == constants.execMode:
        # validate the model and generate report
        print('Validating model ...')
        model_application.eval_model(test_dir, save_dir, model_name, confidence_step)
