# Standard packages
import os
import random
from datetime import datetime
from itertools import product

# Additional packages
import imageio
import numpy as np
from skimage.util import view_as_windows

# Local modules
from patchGenerationTool import constants
from commonTools import common_defs, utils, wf_oct


# Create subfolders based on label types
def create_subfolders(input_dir, unique_id):
    case_dir = input_dir + '/' + unique_id
    if not os.path.isdir(case_dir):
        os.mkdir(case_dir)

    all_subfolders = common_defs.posSubfolders + common_defs.negSubfolders
    all_subfolders.append(common_defs.randomSubfolder)
    for label_type in all_subfolders:
        label_type_dir = case_dir + '/' + label_type
        if not os.path.isdir(label_type_dir):
            os.mkdir(label_type_dir)


# Set has positive labels
def has_positive_labels(scan_id, label_data):
    for label in label_data:
        if scan_id == label['scanId'] and utils.positive_label(label['type']):
            return True
    return False

# Margin has an Exclude Margin labels
def exclude_margin(scan_id, label_data):
    for label in label_data:
        if scan_id == label['scanId'] and label['type'] == common_defs.LabelType.Exclude_Margin:
            return True
    return False

# Export patches from case list
def export_patches(input_dir, save_dir, case_list):
    for case in case_list:
        unique_id = case['uniqueId']

        print('Extracting patches from case %s ...' % unique_id)

        label_data, ground_truth = utils.read_mls(input_dir, unique_id)
        if not (label_data or ground_truth):
            continue

        for scan_id in case['scanId']:
            scan_id_str = 'S' + str(scan_id).zfill(2)
            scan_ground_truth = 0
            for scan_item in ground_truth:
                if scan_item['scanId'] == scan_id:
                    scan_ground_truth = scan_item['groundTruth']

            print(unique_id, ':', scan_id_str)

            if scan_ground_truth == common_defs.GroundTruth.No_Selection:
                print('Error: scan ground truth is not defined for scan #%d - patches not exported ...' % scan_id)
                continue
            elif scan_ground_truth == common_defs.GroundTruth.Close_DCIS:
                print('Error: scan ground truth is set to Close_DCIS for scan #%d - patches not exported ...' % scan_id)
                continue
            elif scan_ground_truth == common_defs.GroundTruth.Close_IDC:
                print('Error: scan ground truth is set to Close_IDC for scan #%d - patches not exported ...' % scan_id)
                continue
            elif scan_ground_truth == common_defs.GroundTruth.Close_ILC:
                print('Error: scan ground truth is set to Close_ILC for scan #%d - patches not exported ...' % scan_id)
                continue
            elif exclude_margin(scan_id, label_data):
                print('Error: scan labels include an "Exclude Margin" label for scan #%d - patches not exported ...' % scan_id)
                continue

            create_subfolders(save_dir, unique_id)

            scan_has_positive_label = has_positive_labels(scan_id, label_data)

            scan_info_filename = input_dir + '/' + unique_id + '/' + scan_id_str + '/scan.json'
            if not os.path.exists(scan_info_filename):
                print('Error: scan configuration file missing for #%d - patches not exported ...' % scan_id)
                continue

            scan_info = wf_oct.ReadScanInfo(scan_info_filename)

            num_bscans = scan_info.numOctSlices
            pixel_width = scan_info.pixDim[0]
            pixel_height = scan_info.pixDim[2]
            slice_distance_mm = scan_info.pixDim[1]
            patch_size_x = int(constants.patchSizeMm / pixel_width + 0.5)
            patch_size_y = int(constants.patchSizeMm / pixel_height + 0.5)
            stride_x = int(constants.trainingHorizontalStrideMm / pixel_width + 0.5)
            stride_y = int(constants.verticalStrideMm / pixel_height + 0.5)
            offset_y = int(constants.verticalOffsetMm / pixel_height + 0.5)
            x_margin = int(constants.horizontalMarginMm / pixel_width + 0.5)
            y_margin = int(constants.verticalMarginMm / pixel_height + 0.5)


            patches = []
            other_patches = []
            for bscan in range(num_bscans):
                bscan_name = input_dir + '/' + unique_id + '/' + scan_id_str + '/OctData_' + \
                             str(bscan).zfill(3) + '.u16'
                scan_data = wf_oct.ReadU16(bscan_name)
                y_location_mm = bscan * slice_distance_mm
                
                data = scan_data.imageData[
                       offset_y: (constants.numVerticalLevels - 1) * stride_y + offset_y + patch_size_y, :]
                data_patches = view_as_windows(data, (patch_size_y, patch_size_x),
                                               ((constants.numVerticalLevels - 1) * stride_y + offset_y, stride_x))
                nv = data_patches.shape[0]
                nh = data_patches.shape[1]
                for y, x in product(range(nv), range(nh)):
                    xstart = x * stride_x
                    xend = xstart + patch_size_x
                    ystart = y * stride_y + offset_y
                    yend = ystart + patch_size_y
                    patch = data_patches[y][x]
                    labeled_patch = False

                    for label in label_data:
                        # label is in B-scan
                        if scan_id == label['scanId'] and abs(y_location_mm - label['y']) < constants.eps:
                            xl = int(label['x'] / pixel_width + 0.5)
                            # uncompensated units
                            yl = int(label['z'] / pixel_height * scan_data.indexOfRefraction + 0.5)
                            # label is within margins
                            if (xstart + x_margin <= xl) and (xl <= xend - x_margin) and \
                                    (ystart + y_margin <= yl) and (yl <= yend - y_margin):
                                label_folder = utils.get_subfolder_from_label(label)
                                file_name = save_dir + '/' + unique_id + '/' + label_folder + '/' + unique_id + '_' + \
                                    scan_id_str + '_%04d_%04d_%04d.png' % (bscan, xstart, ystart)
                                if scan_ground_truth == common_defs.GroundTruth.Negative and utils.positive_label(label['type']) and scan_has_positive_label:
                                    file_name = save_dir + '/' + unique_id + '/' + label_folder + '/' + unique_id + '_' + \
                                                scan_id_str + '_%04d_%04d_%04d' % (bscan, xstart, ystart) + constants.false_positive_tag + '.png'

                                if filename_exists(patches, file_name):
                                    print('Duplicate patch %s for label id %s, skipping' % (file_name, label['labelId']))
                                    continue
                                else:
                                    patches.append([file_name, patch])

                                labeled_patch = True
                                # flip positive patch to double the number
                                if scan_ground_truth > common_defs.GroundTruth.Negative:
                                    file_name_lr = save_dir + '/' + unique_id + '/' + label_folder + '/' + unique_id + \
                                               '_' + scan_id_str + '_%04d_%04d_%04d' % (bscan, xstart, ystart) + constants.flipped_patch_tag + '.png'
                                    patch_lr = np.fliplr(patch)
                                    patches.append([file_name_lr, patch_lr])
                                elif scan_ground_truth == common_defs.GroundTruth.Negative and utils.positive_label(label['type']) and scan_has_positive_label:
                                    file_name_lr_fp = save_dir + '/' + unique_id + '/' + label_folder + '/' + unique_id + \
                                                   '_' + scan_id_str + '_%04d_%04d_%04d' % (bscan, xstart, ystart) + constants.false_positive_tag + constants.flipped_patch_tag + '.png'
                                    patch_lr_fp = np.fliplr(patch)
                                    patches.append([file_name_lr_fp, patch_lr_fp])


                    if not labeled_patch and scan_ground_truth == common_defs.GroundTruth.Negative:
                        label_folder = common_defs.randomSubfolder
                        file_name = save_dir + '/' + unique_id + '/' + label_folder + '/' + unique_id + '_' + \
                            scan_id_str + '_%04d_%04d_%04d_rand.png' % (bscan, xstart, ystart)
                        other_patches.append([file_name, patch])

            # delete random patches to leave N per scan but only for scans with no "false positives"
            if scan_ground_truth == common_defs.GroundTruth.Negative and not scan_has_positive_label:
                add_random_patches(patches, other_patches)

            save_patches(patches)
            print('%d patches saved' % len(patches))


# Add 2000 randomly selected patches to patch list
def add_random_patches(patches, other_patches):
    num_random_patches = 0
    random.seed(datetime.now())
    while num_random_patches < constants.numRandomPatches:
        patch_to_delete = random.randrange(0, len(other_patches))
        patches.append(other_patches[patch_to_delete])
        del other_patches[patch_to_delete]
        num_random_patches += 1


# Save patches as PNG files
def save_patches(patch_list):
    for file_name, patch in patch_list:
        imageio.imwrite(file_name, patch)

def filename_exists(patch_list, file_name_check):
    for file_name, _ in patch_list:
        if file_name == file_name_check:
            return True
    return False