# Standard packages
import glob
import os
import operator
from functools import reduce
import json
from datetime import datetime
import shutil
import csv
# Additional packages
from imageio import imwrite
import numpy as np
from skimage.util import view_as_windows
from sklearn.cluster import DBSCAN

# Local modules
from commonTools import common_defs, utils, wf_oct
from processingTool import clusters, constants, imgassist_processor, stats


# Process case list by ImgAssist
def process_scans(net, model_timestamp, confidence, input_dir, case_list, save_dir, thumbnail_algorithm):
    stats.total_margins = 0
    stats.pos_margins = 0
    stats.neg_margins = 0
    stats.margins_with_pos_clusters = 0
    stats.margins_with_pos_centers = 0
    stats.total_true_positives = 0
    stats.total_true_negatives = 0
    stats.total_false_positives = 0
    stats.total_false_negatives = 0
    stats.total_per_pos = 0
    stats.total_per_neg = 0
    stats.total_clusters_per_pos = 0
    stats.total_clusters_per_neg = 0
    stats.total_pos_clusters = 0
    stats.total_pos_centers = 0
    stats.total_skipped_pos = 0
    stats.total_skipped_neg = 0

    mls_available = False

    date_time = datetime.now()
    out_filename = save_dir + '/' + 'processing_log_' + date_time.strftime('%m%d%y_%H%M%S') + '.txt'
    summary_filename = save_dir + '/' + 'model_summary.csv'
    margin_filename = save_dir + '/' + 'margin_' + date_time.strftime('%m%d%y_%H%M%S') + '.csv'

    fieldname_margin = ["unique ID","Scan ID","Ground Truth","Number of suspicious MLS labels","B-Scan processed",
                 "Total patches","TP patches","FN patches","TN patches","FP patches","Discarded positive patches",
                 "Discarded negative patches","Total cluster","Cluster with true positive patches",
                 "True positive cluster with thumbnails","Elapsed time"]
    with open(margin_filename, 'w', newline='') as f:
        writer = csv.writer(f)
        writer.writerow(fieldname_margin)
    print('Processing case list ...')
    for case in case_list:
        unique_id = case['uniqueId']
        # Unique ID Format is XXXXXXX_PYYYYYY, where YYYYYY is an integer corresponding to case id
        case_id_str = unique_id[-6:]
        case_id = int(case_id_str)
        case_mls_available = False
        label_data, ground_truth = utils.read_mls(input_dir, unique_id)
        if label_data:
            case_mls_available = True
            mls_available = True

        for scan_id in case['scanId']:
            scan_id_str = 'S' + str(scan_id).zfill(2)
            print(unique_id, ':', scan_id_str)

            # number of available suspicious MLS labels per scan
            num_labels = 0
            for label in label_data:
                if label['scanId'] == scan_id and utils.positive_label(label['type']):
                    num_labels += 1

            scan_status = 0
            valid_scan = True
            error_msg = ''
            for status in ground_truth:
                if status['scanId'] == scan_id:
                    scan_status = status['groundTruth']
            if case_mls_available and scan_status == common_defs.GroundTruth.No_Selection:
                error_msg = 'Error: margins with undefined ground truth are ignored'
                print(error_msg)
                valid_scan = False
            elif scan_status == common_defs.GroundTruth.Negative and num_labels > 0:
                error_msg = 'Error: negative margins containing suspicious labels are ignored'
                print(error_msg)
                valid_scan = False
            elif scan_status > common_defs.GroundTruth.Negative and num_labels == 0:
                error_msg = 'Error: positive margins with no suspicious labels are ignored'
                print(error_msg)
                valid_scan = False
            elif scan_status == common_defs.GroundTruth.Close_IDC or scan_status == common_defs.GroundTruth.Close_DCIS or scan_status == common_defs.GroundTruth.Close_ILC:
                error_msg = 'Error: margins with ground truth set to Close_IDC, CLose_DCIS or Close_ILC are ignored'
                print(error_msg)
                valid_scan = False

            scan_folder = input_dir + '/' + unique_id + '/' + scan_id_str
            if not os.path.exists(scan_folder):
                error_msg = 'Error: scan folder %s not found ...' % scan_folder
                print(error_msg)
                valid_scan = False

            scan_info_filename = input_dir + '/' + unique_id + '/' + scan_id_str + '/scan.json'
            if not os.path.exists(scan_info_filename):
                error_msg = 'Error: configuration file missing for scan #%d - margin not processed' % scan_id
                print(error_msg)
                valid_scan = False

            if valid_scan:
                stats.total_margins += 1
                if scan_status > common_defs.GroundTruth.Negative:
                    stats.pos_margins += 1
                else:
                    stats.neg_margins += 1

                scan_info = wf_oct.ReadScanInfo(scan_info_filename)

                num_bscans = scan_info.numOctSlices
                pixel_width = scan_info.pixDim[0]
                pixel_height = scan_info.pixDim[2]
                slice_distance_mm = scan_info.pixDim[1]
                patch_size_x = int(constants.patchSizeMm / pixel_width + 0.5)
                patch_size_y = int(constants.patchSizeMm / pixel_height + 0.5)
                stride_x = int(constants.horizontalStrideMm / pixel_width + 0.5)
                offset_y = int(constants.verticalOffsetMm / pixel_height + 0.5)

                num_patches = 0
                num_detections = 0
                num_clusters = 0
                detections = []
                patches = []
                patch_names = []

                stats.num_true_positives = 0
                stats.num_true_negatives = 0
                stats.num_false_positives = 0
                stats.num_false_negatives = 0
                stats.num_true_pos_clusters = 0
                stats.num_true_pos_thumbnails = 0

                start_time = datetime.now()

                for bscan in range(num_bscans):
                    bscan_name = input_dir + '/' + unique_id + '/' + scan_id_str + '/OctData_' + \
                                 str(bscan).zfill(3) + '.u16'
                    scan_data = wf_oct.ReadU16(bscan_name)
                    y_location_mm = bscan * slice_distance_mm
                    data = scan_data.imageData[offset_y: offset_y + patch_size_y, :]
                    patch_data = view_as_windows(data, (patch_size_y, patch_size_x), (offset_y, stride_x))

                    num_bscan_patches = patch_data.shape[1]
                    num_patches += num_bscan_patches

                    labels = set_patch_labels(scan_id, scan_data, y_location_mm, label_data, num_bscan_patches,
                                              patch_size_x, patch_size_y, stride_x, offset_y, pixel_width, pixel_height)
                    bscan_detections, bscan_patches, bscan_patch_names = \
                        imgassist_processor.analyze_data(net, confidence, patch_data, labels, bscan, y_location_mm,
                                                         pixel_width, stride_x, offset_y, scan_data.indexOfRefraction)

                    if len(bscan_detections) > 0:
                        detections.append(bscan_detections)
                        patches.append(bscan_patches)
                        patch_names.append(bscan_patch_names)

                if detections:
                    # flatten list to produce list of detections per scan
                    detections = reduce(operator.add, detections)
                    num_detections = len(detections)
                    patches = reduce(operator.add, patches)
                    patch_names = reduce(operator.add, patch_names)
                    # ensure clustering threshold is slightly larger than inter-slice distance
                    clusters, cluster_patches, cluster_patch_names, cluster_list = \
                        extract_clusters(detections, patches, patch_names, slice_distance_mm + 0.01, 2, thumbnail_algorithm)
                    num_clusters = len(clusters)

                    elapsed_time = datetime.now() - start_time

                    # sorting generated clusters in ascending order
                    y_coords = [row[1] for row in clusters]
                    indices_sorted = sorted(range(len(y_coords)), key=lambda x: y_coords[x])
                    cluster_patches = [x for _, x in sorted(zip(indices_sorted, cluster_patches))]
                    cluster_patch_names = [x for _, x in sorted(zip(indices_sorted, cluster_patch_names))]

                    save_thumbnails(unique_id, scan_id, cluster_patches, cluster_patch_names, save_dir)
                    save_results(unique_id, case_id, scan_id, model_timestamp, confidence, cluster_list, input_dir, save_dir)
                    save_log(unique_id, scan_id_str, scan_status, case_mls_available, num_labels, num_bscans, num_patches,
                             num_clusters, elapsed_time, out_filename)
                    csv_save(case_mls_available,scan_status,num_clusters, unique_id, scan_id_str, num_bscans, num_patches, elapsed_time, num_labels, margin_filename)

                    stats.total_true_positives += stats.num_true_positives
                    stats.total_true_negatives += stats.num_true_negatives
                    stats.total_false_positives += stats.num_false_positives
                    stats.total_false_negatives += stats.num_false_negatives

                    if stats.num_true_pos_clusters > 0:
                        stats.total_pos_clusters += stats.num_true_pos_clusters
                        stats.margins_with_pos_clusters += 1
                    if stats.num_true_pos_thumbnails > 0:
                        stats.total_pos_centers += stats.num_true_pos_thumbnails
                        stats.margins_with_pos_centers += 1

                if scan_status == common_defs.GroundTruth.Negative:
                    stats.total_per_neg += num_detections
                    stats.total_clusters_per_neg += num_clusters
                else:
                    stats.total_per_pos += num_detections
                    stats.total_clusters_per_pos += num_clusters
            else:
                save_log_error(unique_id, scan_id_str, scan_status, error_msg, out_filename)


    if stats.total_margins > 0:
        save_summary(model_timestamp, confidence, date_time, mls_available, out_filename, thumbnail_algorithm)
        save_csv_summary(model_timestamp, confidence, date_time,summary_filename, thumbnail_algorithm)


# Label patches as suspicious/non-suspicious
def set_patch_labels(scan_id, scan_data, y_location_mm, label_data, num_bscan_patches, patch_size_x, patch_size_y,
                     stride_x, offset_y, pixel_width, pixel_height):
    x_margin = int(constants.horizontalMarginMm / pixel_width + 0.5)
    y_margin = int(constants.verticalMarginMm / pixel_height + 0.5)
    labels = np.zeros(num_bscan_patches, dtype=np.uint8)
    if label_data:
        for n in range(num_bscan_patches):
            x_start = n * stride_x
            y_start = offset_y
            x_end = x_start + patch_size_x
            y_end = y_start + patch_size_y

            for label in label_data:
                # label is in bscan
                if scan_id == label['scanId'] and abs(y_location_mm - label['y']) < constants.eps:
                    xl = int(label['x'] / pixel_width + 0.5)
                    # uncompensated units
                    yl = int(label['z'] / pixel_height * scan_data.indexOfRefraction + 0.5)
                    # label is within margins
                    if (x_start + x_margin <= xl) and (xl <= x_end - x_margin) and \
                       (y_start + y_margin <= yl) and (yl <= y_end - y_margin) and \
                       label['type'] <= common_defs.LabelType.IDC:
                        labels[n] = label['type'] + 1
    return labels


# Extract clusters from list of detections
def extract_clusters(detections, patches, patch_names, eps, min_samples, thumbnail_algorithm):
    coords = np.asarray(detections)[:, :3]
    conf = np.asarray(detections)[:, 4:].ravel()
    ground_truth = np.asarray(detections)[:, 3:4]
    db = DBSCAN(eps=eps, min_samples=min_samples).fit(coords)
    lb = db.labels_
    labels = set(lb[lb >= 0])

    skipped_ground_truth = ground_truth[np.where(lb < 0)].ravel()
    stats.num_skipped_pos = np.count_nonzero(skipped_ground_truth)
    stats.num_skipped_neg = skipped_ground_truth.size - stats.num_skipped_pos
    stats.total_skipped_pos += stats.num_skipped_pos
    stats.total_skipped_neg += stats.num_skipped_neg

    cluster_coords_list = []
    cluster_thumbnails_list = []
    cluster_names_list = []
    cluster_list = []
    
    for lb_id in labels:
        cluster_coords = coords[np.where(lb == lb_id)]
        cluster_size = cluster_coords.size // 3
        cluster_patches = np.asarray(patches)[np.where(lb == lb_id)]
        cluster_names = np.asarray(patch_names)[np.where(lb == lb_id)]
        cluster_conf = conf[np.where(lb == lb_id)]
        cluster_ground_truth = ground_truth[np.where(lb == lb_id)].ravel()
        cluster_label = np.amax(cluster_ground_truth)

        key_img, cluster_score = calculate_key_image_index_and_score(cluster_conf, thumbnail_algorithm, cluster_size)
        thumbnail_coords = cluster_coords[key_img]
        thumbnail_rgb = np.zeros((constants.thumbnailSize, constants.thumbnailSize, 3), dtype=np.uint8)

        for k in range(3):
            thumbnail_rgb[:, :, k] = cluster_patches[key_img]
        thumbnail_label = cluster_ground_truth[key_img]
        thumbnail_name = cluster_names[key_img]
        
        if cluster_label > 0:
            stats.num_true_pos_clusters += 1
            if thumbnail_label > 0:
                thumbnail_rgb = paint_thumbnail_border(thumbnail_rgb, constants.greenColor)
                stats.num_true_pos_thumbnails += 1
            else:
                thumbnail_rgb = paint_thumbnail_border(thumbnail_rgb, constants.redColor)
        
        # Cluster location is specified based on center of cluster
        midp = cluster_size // 2
        cluster_width = cluster_coords[cluster_size-1][1] - cluster_coords[0][1]
        cluster = clusters.Cluster(cluster_coords[midp][0], cluster_coords[midp][1], cluster_coords[midp][2], constants.patchSizeMm, cluster_width, cluster_size, cluster_score)
        cluster.thumbnail = clusters.ClusterThumbnail(thumbnail_coords[0], thumbnail_coords[1], thumbnail_coords[2], constants.patchSizeMm, constants.patchSizeMm)
        cluster_list.append(cluster)
        cluster_coords_list.append([thumbnail_coords[0], thumbnail_coords[1], thumbnail_coords[2], cluster_size, cluster_score])
        cluster_thumbnails_list.append(thumbnail_rgb)
        cluster_names_list.append(thumbnail_name)
        
    return cluster_coords_list, cluster_thumbnails_list, cluster_names_list, cluster_list


# Paint border of thumbnail
def paint_thumbnail_border(thumbnail_rgb, color):
    for k in range(3):
        thumbnail_rgb[0:constants.thumbnailSize, 0:2, k] = color[k]
        thumbnail_rgb[0:constants.thumbnailSize, constants.thumbnailSize - 2:constants.thumbnailSize, k] = color[k]
        thumbnail_rgb[0:2, 0:constants.thumbnailSize, k] = color[k]
        thumbnail_rgb[constants.thumbnailSize - 2:constants.thumbnailSize, 0:constants.thumbnailSize, k] = color[k]
    return thumbnail_rgb

# Find key image based on midpoint or moving average max and provide a moving average max confidence score
def calculate_key_image_index_and_score(cluster_conf, thumbnail_algorithm, cluster_size):
    key_img = 0
    highest_moving_avg = 0

    if len(cluster_conf) >= 3: # Use moving average if there are 3 or more patches in a cluster
        moving_avg = np.convolve(cluster_conf,np.ones(3)/3,mode = 'valid') # create a list of moving average
        highest_avg_idx = np.argmax(moving_avg) # Gets index of highest moving average
        highest_moving_avg = np.max(moving_avg)
        key_img = highest_avg_idx + np.argmax(cluster_conf[highest_avg_idx:highest_avg_idx+3]) # Highest moving average cluster max

    else: #Use average if there is less than 3 patches in a cluster
        highest_moving_avg = np.mean(cluster_conf)
        key_img = np.argmax(cluster_conf)

    # If midpoint is used, overwrite selected key image but leave highest moving average the same
    if thumbnail_algorithm == constants.thumbnail_algorithm_midpoint:
        key_img = cluster_size // 2
    
    return key_img, highest_moving_avg



# Save clusters in a JSON file
def save_results(unique_id, case_id, scan_id, model_timestamp, confidence, clusters, input_dir, save_dir):
    scan_status = 0
    if len(clusters) > 0:
        scan_status = 1
    serial_number = unique_id
    results = {}
    results.update({"caseID": case_id})
    results.update({"scanID": scan_id})
    results.update({"serialNumber": serial_number})
    results.update({"scanStatus": scan_status})
    results.update({"imgAssistHighConfidence": confidence})
    results.update({"imgAssistLowConfidence": 0.0})
    results.update({"imgAssistVersion": model_timestamp})
    results.update({"fileVersion": constants.jsonFileVersion})
    patch_size = {}
    patch_size.update({"height": constants.patchSizeMm, "width": constants.patchSizeMm})
    results.update({"patchSizeMm": patch_size})
    cluster_data = {"clusters": []}
    for i in range(len(clusters)):
        # Cluster coordinates is the top left corner of the center of the cluster
        clusterCoord = {}
        clusterCoord.update({"x": clusters[i].coordX, "y": clusters[i].coordY, "z":clusters[i].coordZ})
        clusterSize = {}
        # Height is from the first b-scan to the last, so we don't count the "thickness" of the last b-scan, hence -1
        clusterSize.update({"width": clusters[i].width, "height":clusters[i].height})
        thumbnail = {}
        thumbnailCoord = {}
        thumbnailCoord.update({"x": clusters[i].thumbnail.coordX, "y": clusters[i].thumbnail.coordY, "z":clusters[i].thumbnail.coordZ})
        thumbnailSize = {}
        thumbnailSize.update({"width": clusters[i].thumbnail.width, "height": clusters[i].thumbnail.height})
        thumbnail.update({"coordMm": thumbnailCoord, "sizeMm": thumbnailSize})
        cluster_data["clusters"].append({"coordMm": clusterCoord, "selected": False, "numberOfDetections": clusters[i].numDetections, "sizeMm": clusterSize, "score": clusters[i].score, "thumbnail": thumbnail})

    results.update(cluster_data)
    out_filepath = save_dir + '/' + unique_id
    if not os.path.isdir(out_filepath):
        os.mkdir(out_filepath)
    out_filepath += '/S' + str(scan_id).zfill(2)
    if not os.path.isdir(out_filepath):
        os.mkdir(out_filepath)
    out_filename = out_filepath + '/imas.json'
    with open(out_filename, 'w') as f:
        json.dump(results, f, indent=4, sort_keys=True)
    out_filepath2 = input_dir + '/' + unique_id + '/S' + str(scan_id).zfill(2)
    out_filename2 = out_filepath2 + '/imas.json'
    with open(out_filename2, 'w') as f:
        json.dump(results, f, indent=4, sort_keys=True)


# Save thumbnails in a PNG file
def save_thumbnails(unique_id, scan_id, cluster_patches, cluster_patch_names, save_dir):
    out_filepath = save_dir + '/' + unique_id
    if not os.path.isdir(out_filepath):
        os.mkdir(out_filepath)
    out_filepath += '/S' + str(scan_id).zfill(2)
    if os.path.isdir(out_filepath):
        shutil.rmtree(out_filepath)
    os.mkdir(out_filepath)
    for i in range(len(cluster_patches)):
        out_image = cluster_patches[i].astype(np.uint8)
        out_filename = out_filepath + '/' + cluster_patch_names[i] + '.png'
        imwrite(out_filename, out_image)
    if len(cluster_patches) > 0:
        montage_cmd = 'montage -mode concatenate -tile 20x -geometry +2+2 -label \'%t\' '
        montage_args = '%s/*_*_*.png %s/imas_clusters.png' % (out_filepath, out_filepath)
        os.system(montage_cmd + montage_args)
        tiles = glob.glob(os.path.join(out_filepath, '*_*_*.png'))
        for f in tiles:
            os.remove(f)


# Save log file with processing info
def save_log(unique_id, scan_id_str, scan_status, case_mls_available, num_labels, num_bscans, num_patches, num_clusters,
             elapsed_time, out_filename):
    ground_truth = utils.ground_truth_to_string(scan_status)

    with open(out_filename, 'a') as f:
        f.write('-'*60)
        f.write('\nScan: ' + unique_id + ' ' + scan_id_str)
        f.write('\nGround truth: ' + ground_truth)
        if case_mls_available:
            if num_labels > 0:
                f.write('\nNumber of suspicious MLS labels: %d' % num_labels)
        else:
            f.write('\nWarning: MLS is not available')
        f.write('\n\nB-scans processed: %d' % num_bscans)
        f.write('\nPatches processed: %d\n' % num_patches)

        if case_mls_available:
            num_negatives = stats.num_true_negatives + stats.num_false_positives
            if scan_status > common_defs.GroundTruth.Negative and (stats.num_true_positives > 0 or
                                                                   stats.num_false_negatives > 0):
                num_positives = stats.num_true_positives + stats.num_false_negatives
                f.write('\nNumber of true positive patches: %d (%3.1f%%)' %
                        (stats.num_true_positives, 100. * stats.num_true_positives / num_positives))
                f.write('\nNumber of false negative patches: %d (%3.1f%%)' %
                        (stats.num_false_negatives, 100. * stats.num_false_negatives / num_positives))
            f.write('\nNumber of true negative patches: %d (%3.1f%%)' %
                    (stats.num_true_negatives, 100. * stats.num_true_negatives / num_negatives))
            f.write('\nNumber of false positive patches: %d (%3.1f%%)\n\n' %
                    (stats.num_false_positives, 100. * stats.num_false_positives / num_negatives))
            if scan_status > common_defs.GroundTruth.Negative:
                f.write('Number of discarded single positive patches: %d\n' % stats.num_skipped_pos)
            f.write('Number of discarded single negative patches: %d' % stats.num_skipped_neg)
            f.write('\n\nNumber of clusters: %d' % num_clusters)
            if scan_status > common_defs.GroundTruth.Negative and num_clusters > 0:
                f.write('\nNumber of clusters with true positive patches: %d' % stats.num_true_pos_clusters)
                if stats.num_true_pos_clusters > 0:
                    f.write('\nNumber of true positive cluster key thumbnails: %d (%3.1f%%)' %
                            (stats.num_true_pos_thumbnails,
                                100. * stats.num_true_pos_thumbnails / stats.num_true_pos_clusters))


        f.write('\n\nElapsed time: %d.%d sec\n' % (elapsed_time.seconds, elapsed_time.microseconds / 1000))
        f.write('-' * 60)
        f.write('\n\n')

def csv_save(case_mls_available,scan_status,num_clusters,unique_id, scan_id_str, num_bscans, num_patches,elapsed_time,num_labels,margin_filename):

    tp, fn, tn, fp, skipped_p, skipped_n, cluster, true_cluster, true_cluster_with_thumbnail = "N/A", "N/A", "N/A", "N/A", "N/A", "N/A", "N/A", "N/A", "N/A"
    if case_mls_available:
        num_negatives = stats.num_true_negatives + stats.num_false_positives
        if scan_status > common_defs.GroundTruth.Negative and (stats.num_true_positives > 0 or
                                                               stats.num_false_negatives > 0):
            num_positives = stats.num_true_positives + stats.num_false_negatives
            tp = '%d (%3.1f%%)' % (stats.num_true_positives, 100. * stats.num_true_positives / num_positives)
            fn = '%d (%3.1f%%)' % (stats.num_false_negatives, 100. * stats.num_false_negatives / num_positives)
        tn = '%d (%3.1f%%)' % (stats.num_true_negatives, 100. * stats.num_true_negatives / num_negatives)
        fp = '%d (%3.1f%%)' % (stats.num_false_positives, 100. * stats.num_false_positives / num_negatives)
        if scan_status > common_defs.GroundTruth.Negative: skipped_p = stats.num_skipped_pos
        skipped_n = stats.num_skipped_neg
        cluster = num_clusters
        if scan_status > common_defs.GroundTruth.Negative and num_clusters > 0:
            true_cluster = stats.num_true_pos_clusters
            if stats.num_true_pos_clusters > 0:
                true_cluster_with_thumbnail = str(stats.num_true_pos_thumbnails) + "(" + str(
                    round(100. * stats.num_true_pos_thumbnails / stats.num_true_pos_clusters, 3)) + "%)"
    else:
        num_labels = "Warning: MLS is not available"

    data_value = [unique_id, scan_id_str, scan_status, num_labels, num_bscans, num_patches,
                  tp, fn, tn, fp, skipped_p, skipped_n, cluster, true_cluster, true_cluster_with_thumbnail,
                  '%d.%d sec' % (elapsed_time.seconds, elapsed_time.microseconds / 1000)]

    with open(margin_filename, 'a') as f:
        writer = csv.writer(f)
        writer.writerow(data_value)


# Save error message to log file
def save_log_error(unique_id, scan_id_str, scan_status, error_msg, out_filename):
    ground_truth = utils.ground_truth_to_string(scan_status)

    with open(out_filename, 'a') as f:
        f.write('-' * 60)
        f.write('\nScan: ' + unique_id + ' ' + scan_id_str)
        f.write('\nGround truth: ' + ground_truth)
        f.write('\n\n%s\n\n' % error_msg)
        f.write('-' * 60)
        f.write('\n')


# Save summary
def save_summary(model_timestamp, confidence, date_time, mls_available, out_filename, thumbnail_algorithm):
    total_positives = stats.total_true_positives + stats.total_false_negatives
    total_negatives = stats.total_true_negatives + stats.total_false_positives
    if total_positives > 0:
        truePositivePercent = 100. * stats.total_true_positives / total_positives
        falseNegativePercent = 100. * stats.total_false_negatives / total_positives
    if total_negatives > 0:
        trueNegativePercent = 100. * stats.total_true_negatives / total_negatives
        falsePositivePercent = 100. * stats.total_false_positives / total_negatives
    aver_per_pos = 0
    aver_per_neg = 0
    if stats.pos_margins > 0:
        aver_per_pos = int(stats.total_per_pos / stats.pos_margins + 0.5)
    if stats.neg_margins > 0:
        aver_per_neg = int(stats.total_per_neg / stats.neg_margins + 0.5)
    aver_per_margin = int((stats.total_per_pos + stats.total_per_neg) / stats.total_margins + 0.5)
    aver_clusters_per_pos = 0
    aver_clusters_per_neg = 0
    if stats.pos_margins > 0:
        aver_clusters_per_pos = int(stats.total_clusters_per_pos / stats.pos_margins + 0.5)
    if stats.neg_margins > 0:
        aver_clusters_per_neg = int(stats.total_clusters_per_neg / stats.neg_margins + 0.5)
    aver_clusters_per_margin = int((stats.total_clusters_per_pos + stats.total_clusters_per_neg) / stats.total_margins + 0.5)

    with open(out_filename, 'r+') as f:
        content = f.read()
        f.seek(0, 0)
        f.write('Test run: ' + date_time.strftime("%Y-%m-%d %H:%M:%S"))
        f.write('\nModel used: ' + model_timestamp)
        f.write('\nThumbnail Algorithm: ' + thumbnail_algorithm)
        f.write('\nConfidence value: ' + str(confidence) + '\n\n')
        f.write('-' * 60)
        f.write('\nSummary:')
        f.write('\n\nNumber of margins evaluated: %d' % stats.total_margins)
        if mls_available:
            f.write('\nNumber of positive margins evaluated: %d\n' % stats.pos_margins)
            if stats.pos_margins > 0:
                f.write('Number of positive margins with positive clusters: %d' % stats.margins_with_pos_clusters)
                f.write('\nNumber of positive margins with positive key thumbnails: %d' % stats.margins_with_pos_centers)
            if total_positives > 0:
                f.write('\n\nNumber of true positive patches: %d (%3.1f%%)' % (stats.total_true_positives, truePositivePercent))
                f.write('\nNumber of false negative patches: %d (%3.1f%%)' % (stats.total_false_negatives, falseNegativePercent))
            if total_negatives > 0:
                f.write('\nNumber of true negative patches: %d (%3.1f%%)' % (stats.total_true_negatives, trueNegativePercent))
                f.write('\nNumber of false positive patches: %d (%3.1f%%)' % (stats.total_false_positives, falsePositivePercent))
            f.write('\n\nAverage number of positive patches per margin: %d' % aver_per_margin)
            if stats.pos_margins > 0:
                f.write('\nAverage number of positive patches per positive margin: %d' % aver_per_pos)
            if stats.neg_margins > 0:
                f.write('\nAverage number of positive patches per negative margin: %d' % aver_per_neg)
            f.write('\n\nNumber of discarded single true positive patches: %d' % stats.total_skipped_pos)
            f.write('\nNumber of discarded single true negative patches: %d' % stats.total_skipped_neg)
            f.write('\n\nNumber of clusters: %d' % (stats.total_clusters_per_pos + stats.total_clusters_per_neg))
            if stats.pos_margins > 0:
                f.write('\nNumber of clusters with true positive patches: %d' % stats.total_pos_clusters)
                f.write('\nNumber of true positive key thumbnails: %d' % stats.total_pos_centers)
            f.write('\n\nAverage number of clusters per margin: %d' % aver_clusters_per_margin)
            if stats.pos_margins > 0:
                f.write('\nAverage number of clusters per positive margin: %d' % aver_clusters_per_pos)
            if stats.neg_margins > 0:
                f.write('\nAverage number of clusters per negative margin: %d' % aver_clusters_per_neg)
        f.write('\n' + '-' * 60 + '\n\n')
        f.write(content)

def save_csv_summary(model_timestamp, confidence, date_time, summary_filename, thumbnail_algorithm):
    total_positives = stats.total_true_positives + stats.total_false_negatives
    total_negatives = stats.total_true_negatives + stats.total_false_positives
    aver_per_pos = 0
    aver_per_neg = 0
    if stats.pos_margins > 0:
        aver_per_pos = int(stats.total_per_pos / stats.pos_margins + 0.5)
    if stats.neg_margins > 0:
        aver_per_neg = int(stats.total_per_neg / stats.neg_margins + 0.5)
    aver_per_margin = int((stats.total_per_pos + stats.total_per_neg) / stats.total_margins + 0.5)
    aver_clusters_per_pos = 0
    aver_clusters_per_neg = 0
    if stats.pos_margins > 0:
        aver_clusters_per_pos = int(stats.total_clusters_per_pos / stats.pos_margins + 0.5)
    if stats.neg_margins > 0:
        aver_clusters_per_neg = int(stats.total_clusters_per_neg / stats.neg_margins + 0.5)
    aver_clusters_per_margin = int(
        (stats.total_clusters_per_pos + stats.total_clusters_per_neg) / stats.total_margins + 0.5)

    with open(summary_filename, 'w', newline='') as f:
        fieldnames = ["Test run", "Model used", "Thumbnail Algorithm", "Confidence value", "Margins",
                      "Positive margins", "Positive margins with positive clusters",
                      "Positive margins with positive key thumbnails", "True positive patches",
                      "False negative patches", "True negative patches",
                      "False positive patches", "Average positive patches per margin",
                      "Average positive patches per positive margin",
                      "Average positive patches per negative margin",
                      "Discarded single true positive patches", "Discarded single true negative patches",
                      "Total clusters", "Clusters with true positive patches",
                      "True positive key thumbnails", "Average clusters per margin",
                      "Average clusters per positive margin", "Average clusters per negative margin"]

        truePositivePercent = 0
        falseNegativePercent = 0
        if total_positives > 0:
            truePositivePercent = 100. * stats.total_true_positives / total_positives
            falseNegativePercent = 100. * stats.total_false_negatives / total_positives

        trueNegativePercent = 0
        falsePositivePercent = 0
        if total_negatives > 0:
            trueNegativePercent = 100. * stats.total_true_negatives / total_negatives
            falsePositivePercent = 100. * stats.total_false_positives / total_negatives

        values = [date_time.strftime("%Y-%m-%d %H:%M:%S"), model_timestamp, thumbnail_algorithm, str(confidence),
                  stats.total_margins, stats.pos_margins,
                  stats.margins_with_pos_clusters, stats.margins_with_pos_centers,
                  '%d (%3.1f%%)' % (
                  stats.total_true_positives, truePositivePercent),
                  '%d (%3.1f%%)' % (
                  stats.total_false_negatives, falseNegativePercent),
                  '%d (%3.1f%%)' % (
                  stats.total_true_negatives, trueNegativePercent),
                  '%d (%3.1f%%)' % (
                  stats.total_false_positives, falsePositivePercent),
                  aver_per_margin, aver_per_pos, aver_per_neg, stats.total_skipped_pos,
                  stats.total_skipped_neg, stats.total_clusters_per_pos + stats.total_clusters_per_neg,
                  stats.total_pos_clusters, stats.total_pos_centers, aver_clusters_per_margin,
                  aver_clusters_per_pos, aver_clusters_per_neg]
        writer = csv.writer(f)
        writer.writerow(fieldnames)
        writer.writerow(values)

