# Standard packages
import sys
import os
import json

# Local modules
from commonTools import imgassist, utils, parse_csv
from processingTool import constants, data_processor, stats


# Print program usage
def usage(progname, version):
    print('ImgAssist Processing Tool v. %s\n\nUsage: %s %s' % (version, progname, '<input_params.json>'))


# Read input parameters
def read_params(filename):
    global tool_name
    global input_version
    global input_dir
    global save_dir
    global dataset_filename
    global model_dir
    global model_name
    global confidence
    global thumbnail_algorithm

    with open(filename) as f:
        data = json.load(f)

    tool_name = data['tool']
    if tool_name != constants.tool:
        print('Error: input parameters incompatible with tool ...')
        sys.exit(0)

    input_version = data['version']
    if constants.version != input_version:
        print('Error: tool version is not compatible with input ...')
        sys.exit(0)

    input_dir = data['inputFolder']
    if not os.path.exists(input_dir):
        print('Error: input folder not found, exiting ...')
        sys.exit(0)

    save_dir = data['destinationFolder']
    if not os.path.exists(save_dir):
        print('Error: destination folder not found, exiting ...')
        sys.exit(0)

    dataset_filename = data['dataSetFilename']
    if not os.path.exists(dataset_filename):
        print('Error: Dataset file not found, exiting ...')
        sys.exit(0)

    model_dir = data['modelFolder']
    if not os.path.exists(model_dir):
        print('Error: model folder not found, exiting ...')
        sys.exit(0)

    model_name = data['modelFileName']

    confidence = data['classificationConfidence']
    if confidence < 0.5 or confidence > 1.0:
        print('Error: Wrong confidence value, exiting ...')
        sys.exit(0)

    if 'thumbnailAlgorithm' in data:
        thumbnail_algorithm = data['thumbnailAlgorithm']
    else:
        thumbnail_algorithm = constants.thumbnail_algorithm_midpoint

    if thumbnail_algorithm == constants.thumbnail_algorithm_ma_max:
        print('Clustering algorithm chosen: Moving Average Maximum')
    else:
        # Ensure algorithm is set to default and not some unknown value
        thumbnail_algorithm = constants.thumbnail_algorithm_midpoint
        print('Default clustering algorithm chosen: Mid-Point')

###
def main(argc, argv):
    # check input arguments
    if argc != 2:
        usage(os.path.basename(argv[0]), constants.version)
        sys.exit(2)

    # read input parameters
    param_file = argv[1]
    if not os.path.exists(param_file):
        print('Error: input parameter file not found, exiting ...')
        sys.exit(0)
    read_params(param_file)

    stats.init()
    case_list = parse_csv.patch_generation(dataset_filename)
    # read and unzip input cases

    utils.extract_data(input_dir, case_list["caseList"])

    # load ImgAssist model
    net, timestamp = imgassist.load_model(model_dir, model_name)

    # process scans by ImgAssist
    data_processor.process_scans(net, timestamp, confidence, input_dir, case_list["caseList"], save_dir, thumbnail_algorithm)
